#! /usr/bin/perl
# 99-03-26	fixed incorrect folding of assignments
# 99-03-31	now only reads structures if going to use them
#		added usage instructions
#		writes partially assigned crosspeaks to new xpk list
# 99-04-05	fixed problems with HSQC-NOESY-HSQC
# 99-04-13	now properly includes unassigned atoms as possibilities
# 99-08-20	reads entire xpk file at once instead of a line at a time
#		identifies possible mirror peaks
#		identifies peaks assigned as mirror peaks
#		no longer able to write new XPK file
# 99-08-23	only reads BMRB shifts if needed
# 99-08-24	added secondary structure element filter
# 99-08-25	exclude deleted NMRView peaks from analysis
# 99-08-26	fixed some problems with secondary structure filter
# 99-09-02	discovered splice doesn't update $# in secondary structure filter
# 99-09-17	only writes possibilities after each filter if some possibilities were eliminated
#		reintroduced ability to write new XPK file
# 99-09-22	new xpk file now writen all at once, after all crosspeaks examined
# 99-10-07	now reads PDB files with chainID column properly
# 99-10-08	tidied the restraint writing subroutine
# 00-03-31	altered the use of BMRB shifts for unassigned resonances
#		now uses mean  std-dev instead of min to max
# 00-04-06	fixed incorrect labeling of ambigs in out file
# 00-04-07	added default value (2) for number of BMRB standard deviations
# 00-04-10	accounted for aliasing of unassigned resonances
#		if tolerance crosses edge of spectrum looks at other edge for possibilities
# 00-04-26	added flag to use BMRB for unassigned resonances
#		can write unassigned peaks to separate crosspeak file
# 00-04-28	prints distances in comment for unique and preassigned restraints
# 00-05-01	only prints 'Preassigned' in UPL if using assignments to filter
# 00-05-03	can now cope with HSQC-NOESY expts as well as NOESY-HSQC
#		doesn't fold assignments for detect dimension
#		added 3D aromatic NOESY expt
# 00-05-04	changed ignore_list from array to string
#		now always performs test for loss of possibility with shortest mean distance
# 00-05-11	always reads second field as volume in Felix vol file, ignores other fields
#		fixed HSQC-NOESY-HSQC, only does 15N
# 00-05-18	fixed problems with folding NMRView data
# 01-08-16	altered regular expression for reading PDB file
#		tidied distance measuring loops
#		added use FileHandle to flush disk IO buffer
# 03-06-03	changed ignore_list back to array, array searched using grep
# 04-04-27	added new xpk file format; standard NMRview

if ($#ARGV < 0 or $ARGV[0] eq "-help") { die "
SANE: Structure Assisted NOE Evaluation

USAGE:   $0 parameter_file

SANE creates a restraint file from a crosspeak list and a
list of assignments. It can use structures to help reduce
ambiguity. See the documentation for more information.

Brendan Duggan - 2003 June 3
\n"; }

use FileHandle;

$input_parameters = $ARGV[0];
require "$input_parameters";

open (OUT,"> $out_file");
open (UPL,"> $upl_file");
open (AMBIG_MAP,"> $ambig_map");

&INITIALISATION;
&READ_ATOM_MAPPING;
&READ_ASSIGNMENTS;
&READ_STRUCTURES if ($using_distances);
&READ_XPK_DATA;
foreach $xpkno ( sort numerically keys %crosspeak ) {
	$message = "";
	$ignoring = "";
	print "Doing crosspeak #$xpkno\n";
	print OUT "\nCrosspeak #$xpkno at ";
	foreach $dimension (1..$number_dimensions) { printf OUT "%7.3f ", $crosspeak{$xpkno}{'shift'}[$dimension]; }
	print OUT "\n";
	&CHECK_WATER;
	&DETERMINE_BIN;
	print OUT "Volume $crosspeak{$xpkno}{'volume'},  Distance Bin $Rij\n";
	print OUT "Assigned as ";
	foreach $dimension (1..$number_dimensions) { print OUT "$crosspeak{$xpkno}{'ass'}[$dimension] "; }
	print OUT "\n";
	&FIND_MIRRORS;
	&FIND_POSSIBILITIES;
	if ($number_possibilities > 0 and $using_distances) {
		&MEASURE_DISTANCES;
		&FILTER_DISTANCES;
	}
	if ($number_possibilities > 0 and $using_assignments) { 
		&FILTER_ASSIGNMENTS;
	}
	if ($number_possibilities > 0 and $using_secondary) { 
		&FILTER_SECONDARY;
	}
	if ($number_possibilities > 1 and $using_contributions and not $diagonal_or_geminal) {
		&CALC_CONTRIBUTIONS;
		&FILTER_CONTRIBUTIONS;		
	}
	&WRITE_RESTRAINTS;
	if ($new_xpk) { &UPDATE_PEAK_LIST; }
}

close (OUT);
close (UPL);
if ($new_xpk) { &WRITE_NEW_XPK; }
if ($unass_xpk) { &WRITE_UNASS_XPK; }

exit;


sub numerically { $a <=> $b; }

sub INITIALISATION
# Have to do some different things for the different
# types of data.  Don't want to do these for every
# crosspeak so do them here, outside loop.
{
print OUT "SANE : Structure Assisted NOE Evaluation\n";
print OUT "----------------------------------------\n";
EXPERIMENT_TYPE: {
	if ($expt_flag >= 0 and $expt_flag < 10) {
		print OUT "Examining crosspeaks from a 2D NOESY.\n\t$xpk_file\n";
		$number_dimensions = 2;
		last EXPERIMENT_TYPE;
	}
	if ($expt_flag >= 10 and $expt_flag < 20) {
		print OUT "Examining crosspeaks from a 3D 15N NOESY.\n\t$xpk_file\n";
		$number_dimensions = 3;
		if ($heteroA_dim and not $heteroB_dim) {	# if HSQC-NOESY rather than NOESY-HSQC swap dims. then can treat as NOESY-HSQC
			($protonA_dim,$heteroA_dim,$protonB_dim,$heteroB_dim,$heteroA,$heteroB) = ($protonB_dim,$heteroB_dim,$protonA_dim,$heteroA_dim,$heteroB,$heteroA);
		}
		last EXPERIMENT_TYPE;
	}
	if ($expt_flag >= 20 and $expt_flag < 30) {
		print OUT "Examining crosspeaks from a 3D 13C NOESY.\n\t$xpk_file\n";
		$number_dimensions = 3;
		if ($heteroA_dim and not $heteroB_dim) {	# if HSQC-NOESY rather than NOESY-HSQC swap dims. then can treat as NOESY-HSQC
			($protonA_dim,$heteroA_dim,$protonB_dim,$heteroB_dim,$heteroA,$heteroB) = ($protonB_dim,$heteroB_dim,$protonA_dim,$heteroA_dim,$heteroB,$heteroA);
		}
		last EXPERIMENT_TYPE;
	}
	if ($expt_flag >= 30 and $expt_flag < 40) {
		print OUT "Examining crosspeaks from a 3D HSQC-NOESY-HSQC.\n\t$xpk_file\n";
		$number_dimensions = 3;
		if ($protonA_dim and not $protonB_dim) {
			($protonA_dim,$heteroA_dim,$protonB_dim,$heteroB_dim,$heteroA,$heteroB) = ($protonB_dim,$heteroB_dim,$protonA_dim,$heteroA_dim,$heteroB,$heteroA);
		}
		last EXPERIMENT_TYPE;
	}
	if ($expt_flag >= 40 and $expt_flag < 50) {
		print OUT "Examining crosspeaks from a 4D.\n\t$xpk_file\n";
		$number_dimensions = 4;
		last EXPERIMENT_TYPE;
	}
	if ($expt_flag >= 50 and $expt_flag < 60) {
		$number_dimensions = 3;
		print OUT "Examining crosspeaks from some sort of select filter.\n\t$xpk_file\n";
		last EXPERIMENT_TYPE;
	}
	if ($expt_flag >= 60 and $expt_flag < 70) {
		print OUT "Examining crosspeaks from a 3D aromatic NOESY.\n\t$xpk_file\n";
		$number_dimensions = 3;
		if ($heteroA_dim and not $heteroB_dim) {	# if HSQC-NOESY rather than NOESY-HSQC swap dims. then can treat as NOESY-HSQC
			($protonA_dim,$heteroA_dim,$protonB_dim,$heteroB_dim,$heteroA,$heteroB) = ($protonB_dim,$heteroB_dim,$protonA_dim,$heteroA_dim,$heteroB,$heteroA);
		}
		last EXPERIMENT_TYPE;
	}
}
# set up indices for location of information in @possibility
$protonA_dev = $protonA_dim + $number_dimensions;
$heteroA_dev = $heteroA_dim + $number_dimensions;
$protonB_dev = $protonB_dim + $number_dimensions;
$heteroB_dev = $heteroB_dim + $number_dimensions;
$shortest = $number_dimensions * 2 + 1;
$average  = $number_dimensions * 2 + 2;
$diageminal = $number_dimensions * 2 + 3;

$nucleus[$protonA_dim] = "H";
$nucleus[$heteroA_dim] = $heteroA;
$nucleus[$protonB_dim] = "H";
$nucleus[$heteroB_dim] = $heteroB;

if ($data_type eq "Felix") { print OUT "\t$vol_file\n"; }

print OUT "\nPossibilities determined with the following chemical shift tolerances:\n";
foreach $dimension (1..$number_dimensions) {
	print OUT "D$dimension\t $nucleus[$dimension]\t $tolerance[$dimension]\n";
}
print OUT "\n";
if ($using_contributions) { $using_distances = "true"; }
if ($using_distances) { print OUT "Filtering possibilities with $distance_cutoff A distance filter.\n"; }
if ($using_contributions) { print OUT "Filtering possibilities with $contribution_cutoff contribution filter.\n"; }
if ($using_assignments) { print OUT "Filtering possibilities with existing assignments.\n"; }
if ($using_secondary) {
	foreach $element ( keys %sse ) {
		($sse{$element}{'start'}, $sse{$element}{'end'}) = split("-",$sse{$element});
		if ($element =~ "helix" and $sse{$element}{'end'}-$sse{$element}{'start'} > $longest_helix) {
			$longest_helix = $sse{$element}{'end'}-$sse{$element}{'start'};
		}
		if ($element =~ "strand" and $sse{$element}{'end'}-$sse{$element}{'start'} > $longest_strand) {
			$longest_strand = $sse{$element}{'end'}-$sse{$element}{'start'};
		}
	}
	while ( $exclude{'helix'} =~ /([abcdegN]{2})([0-9]{1,2})(\+)/ ) {
		$stem = $1;
		$start = $2;
		$string = "";
		foreach $i ($start..$longest_helix) { $string .= "$stem$i,"; }
		$exclude{'helix'} =~ s/(\w{3}\+)/$string/;
		$exclude{'helix'} =~ s/,,/,/;
	}
	if ( $exclude{'helix'} !~ /,$/ ) { $exclude{'helix'} .= ","; }
	while ( $exclude{'strand'} =~ /([abcdegN]{2})([0-9]{1,2})(\+)/ ) {
		$stem = $1;
		$start = $2;
		$string = "";
		foreach $i ($start..$longest_strand) { $string .= "$stem$i,"; }
		$exclude{'strand'} =~ s/(\w{3}\+)/$string/;
		$exclude{'strand'} =~ s/,,/,/;
	}
	if ( $exclude{'strand'} !~ /,$/ ) { $exclude{'strand'} .= ","; }
	if ( $exclude{'all'} !~ /,$/ ) { $exclude{'all'} .= ","; }
	print "Read secondary structure elements\n";
	print OUT "Excluding from helices $exclude{'helix'}\n";
	print OUT "Excluding from b strands $exclude{'strand'}\n";
	print OUT "Excluding everywhere $exclude{'all'}\n";
}
print OUT "Ignoring peaks with chemical shifts of $water_shift  $water_tolerance ppm.\n\n";

DATA_TYPE: {
	if ( $data_type eq "NMRView" or $data_type eq "tsriNMRView") {
	# For NMRView data have to read a sequence file because 
	# residue names are not included in the peak assignments.
	# The sequence is stored in an array @sequence.
		@sequence[0] = "";	# put nothing in $sequence[0] to keep numbering consistent
		open (SEQ,$seq_file) or warn "Can't open $seq_file";
		while ($line = <SEQ>) {
			chomp($line);
			push(@sequence,uc($line));
		}
		close(SEQ);
		print "Read $#sequence residues from $seq_file\n";
		foreach $dimension (1..$number_dimensions) {
			$swp[$dimension] = ($lower[$dimension]-$upper[$dimension])*($number_points[$dimension]+1)/$number_points[$dimension];
		}
		last DATA_TYPE;
	}
	if ( $data_type eq "Felix" ) {
	# For Felix data have to calculate factors to convert
	# points to ppm.  Also have to determine upper and lower
	# limits in ppm and sweep width of each dimension.  Have
	# to open a volume file because volumes are not stored 
	# with crosspeaks.
		foreach $dimension (1..$number_dimensions) {
			$factor[$dimension] = $spectral_width[$dimension]/($frequency[$dimension]*$number_points[$dimension]);
			$upper[$dimension] = $reference_ppm[$dimension] - ($number_points[$dimension] - $reference_point[$dimension])*$factor[$dimension];
			$lower[$dimension] = $reference_ppm[$dimension] - (1 - $reference_point[$dimension])*$factor[$dimension];
			$swp[$dimension] = $spectral_width[$dimension]/$frequency[$dimension];
		}
		last DATA_TYPE;
	}
}
}

sub READ_STRUCTURES
# Reads ensemble of structures into data structure called atom.
# First subscript is the number of the structure (0..n-1).
# Second subscript is a hash, the name of the atom.  Third subscript 
# is a hash, 'xpos', 'ypos' or 'zpos'.  The atom name is stored 
# in the form AAA_###:aaa, where AAA is the three letter code for 
# amino acid residues in capitals, ### is the residue number, 
# and aaa is the atom name.
{
if ($#ensemble == -1) {
	print "No structures defined. Cannot do distance or contribution filtering\n";
	print OUT "No structures defined. Cannot do distance or contribution filtering\n";
	$using_distances = "";
	$using_contributions = "";
} else {
	$structure_number = 0;
	foreach $structure (@ensemble) {
		$atom_number = 0;
		open (PDB,$structure) or die "Can't open $structure\n";
		while ($line = <PDB>) {
			CHECK: {
				# PDB file without chain ID field
				if ( $line =~ /^ATOM\s+\d+\s+(\w+)\s+([A-Z]+)\s+(\d+)\s+([-\d\.]+)\s+([-\d\.]+)\s+([-\d\.]+)/ ) {
					$name = $2."_".$3.":".$1;
					$atom[$structure_number]{$name}{'xpos'} = $4;
					$atom[$structure_number]{$name}{'ypos'} = $5;
					$atom[$structure_number]{$name}{'zpos'} = $6;
					$atom_number++;
					last CHECK;
				}
				# PDB file with chain ID field
				if ( $line =~ /^ATOM\s+\d+\s+(\w+)\s+([A-Z]+)\s+(\w+)\s+(\d+)\s+([-\d\.]+)\s+([-\d\.]+)\s+([-\d\.]+)/ ) {
					$name = $2."_".$4.":".$1;
					$chainID = $3;
					$atom[$structure_number]{$name}{'xpos'} = $5;
					$atom[$structure_number]{$name}{'ypos'} = $6;
					$atom[$structure_number]{$name}{'zpos'} = $7;
					$atom_number++;
					last CHECK;
				}
			}
		}
		$structure_number++;
		close (PDB);
		print "Read $atom_number atoms in structure $structure_number : $structure\n";
		print OUT "Read $atom_number atoms in structure $structure_number : $structure\n";
	}
}
}

sub READ_ATOM_MAPPING
# Reads an AMBER style map and stores pseudoatom names with 
# the list of their equivalent atoms.  Data structure is called map.
# First subscript is residue name as a hash.  Second subscript 
# is pseudoatom name as a hash.  Each entry is a list of the 
# atoms that the pseudoatom corresponds to.
# NOTE: The map file must contain every atom name used in the
# assignments, even if they are not pseudoatoms.
{
open (MAP,$map_file) or die "Can't open $map_file\n";
while ($line = <MAP>) {
	if (substr($line,0,7) eq "RESIDUE" or substr($line,0,7) eq "MAPPING") {
		@entry = split(/\s+/,$line);
		if ($entry[0] eq "RESIDUE") {
			$residue = $entry[1];
		} elsif ($entry[0] eq "MAPPING") {
			$pseudoatom = $entry[1];
			@individual_atoms = ();
			foreach $i (3..$#entry) {
				push @individual_atoms, $entry[$i];
			}
			$map{$residue}{$pseudoatom} = [ @individual_atoms ];
		}
	}
}
close (MAP);
print "Read atom mapping from $map_file\n";
print OUT "Read atom mapping from $map_file\n";
}

sub READ_STANDARD_SHIFTS
# Reads a file containing the maximum and minimum chemical shifts
# reported for every type of protein atom. These limits are used
# later on to decide if an unassigned atom could be a possibility.
# Lines starting with a '#' are ignored.  Information is stored
# in the hash 'bmrb'. First subscript is the residue name. Second
# subscript is the atom name.  Third subscript is 'maximum' or
# 'minimum'.
{
open (SHIFTS,$standard_shifts) or die "Can't open $standard_shifts\n";
while ($line = <SHIFTS>) {
	while ( substr($line, 0, 1) eq "#" ) { $line = <SHIFTS>; }
	($residue,$atom_name,$nucleus,$number,$min,$max,$mean,$stddev) = split(/\s+/,$line);
	$bmrb{$residue}{$atom_name}[0]{"std dev"} = $stddev;
	foreach $dimension (1..$number_dimensions) {
		$folded_shift = $mean;
		if ( $nucleus eq $nucleus[$dimension] ) {
			if ($dimension != $detect_dim) {
				while ($folded_shift < $upper[$dimension]) { $folded_shift += $swp[$dimension]; }
				while ($folded_shift > $lower[$dimension]) { $folded_shift -= $swp[$dimension]; }
				$bmrb{$residue}{$atom_name}[$dimension]{"mean"} = $folded_shift;
			} else {
				if ($folded_shift >= $upper[$detect_dim] and $folded_shift <= $lower[$detect_dim]) {
					$bmrb{$residue}{$atom_name}[$dimension]{"mean"} = $folded_shift;
				}
			}
		}
	}
}
close (SHIFTS);
print "Read standard chemical shift ranges from $standard_shifts\n";
print OUT "Read standard chemical shift ranges from $standard_shifts\n";
if (not $number_BMRB_stddevs) { $number_BMRB_stddevs = 2; }
print "Tolerance for unassigned resonances is $number_BMRB_stddevs standard deviations\n";
print OUT "Tolerance for unassigned resonances is $number_BMRB_stddevs standard deviations\n";
}

sub READ_ASSIGNMENTS
# Reads assignments from Ishwar style assignment list and 
# stores in data structure assignment. Subscript is the atom 
# name as a hash. Each element stores one particular chemical 
# shift. Atom names are in the form AAA_###:aaa. All atom names 
# are converted to upper case. If chemical shift is less than 
# or equal to -999 then the atom is flagged as unassigned.
# Unassigned atoms must use the names used in the BMRB chemical 
# shift list.
{
open (ASS,$ass_file) or die "Can't open $ass_file\n";
while ($line = <ASS>) {
	$line =~ s/^\s+//;	# remove leading whitespace
	if ( $data_type eq "Felix" ) {
		($residue_number, $residue_name, $atom_name, $chemical_shift) = split(/\s+/,$line);
	} elsif ( $data_type eq "NMRView" or $data_type eq "tsriNMRView") {
		($residue, $chemical_shift, $flag) = split(/\s+/,$line);
		($residue_number, $atom_name) = split(/\./,$residue);
		$residue_name = $sequence[$residue_number];
	}
	foreach $dimension (1..$number_dimensions) {
		$folded_shift = $chemical_shift;
		if ( (substr($atom_name,0,1) eq $nucleus[$dimension]) or
		     (substr($atom_name,0,1) eq "Q" and $nucleus[$dimension] eq "H") or
		     (substr($atom_name,0,1) eq "M" and $nucleus[$dimension] eq "H") or
		     (substr($atom_name,0,1) eq "C" and $nucleus[$dimension] eq "CN") or
		     (substr($atom_name,0,1) eq "N" and $nucleus[$dimension] eq "CN") ) {
			if ($using_unassigned and ($folded_shift <= -999 or $folded_shift >= 999)) {
				$folded_shift = "unassigned";
				$need_BMRB_shifts = "true";
			} else {
				if ($dimension != $detect_dim) {	# fold shifts if not detect dimension
					while ($folded_shift < $upper[$dimension]) { $folded_shift += $swp[$dimension]; }
					while ($folded_shift > $lower[$dimension]) { $folded_shift -= $swp[$dimension]; }
				} else {	# if detect dimension, do not fold and make sure within spectral width
					if ($folded_shift < $upper[$detect_dim] or $folded_shift > $lower[$detect_dim]) { next; }
				}
			}
			$name = uc($residue_name."_".$residue_number.":".$atom_name);
			$assignment[$dimension]{$name} = $folded_shift;
			$number_assignments[$dimension]++;
		}
	}
}
close (ASS);
print "Read assignments from $ass_file\n";
print OUT "Read assignments from $ass_file\n";
foreach $dimension (1..$number_dimensions) {
	print OUT "$number_assignments[$dimension] in D$dimension\t";
}
print OUT "\n";
if ( $need_BMRB_shifts ) { &READ_STANDARD_SHIFTS; }
}

sub READ_XPK_DATA
{
open (XPK,$xpk_file) or die "Can't open $xpk_file\n";				#        to ignore header
if ( $data_type eq "Felix" ) { open (VOL,$vol_file) or die "Can't open $vol_file\n"; $vol_data = <VOL>; $vol_data = <VOL>; }
while ($line = <XPK>) {
	$line =~ s/^\s+//;
	@peak_entry = split(/\s+/,$line);
	SWITCH: {
		if ( $data_type eq "NMRView" ) {
			while ( $#peak_entry != $number_dimensions*7 + 5 ) {
				if ($new_xpk) { $new_xpk_header .= $line; }	# read through header
				$line = <XPK>;					# storing lines in string
				$line =~ s/^\s+//;				# to print in new_xpk file
				@peak_entry = split(/\s+/,$line);	# and splitting the next line
			}
            $xpkno = $peak_entry[0];
            if ($new_xpk or $unass_xpk) { $xpk_data{$xpkno} = [ @peak_entry ]; }
            if ($peak_entry[$number_dimensions*7 + 5] != -1) {  # make sure peak has not been marked for deletion
                foreach $dimension (1..$number_dimensions) {
                    $crosspeak{$xpkno}{'shift'}[$dimension] = $peak_entry[$dimension*7 - 5];
                    ($number, $atom) = ($peak_entry[$dimension*7 - 6] =~ /^\{(\d+)\.(\S+)\}$/);
                    if ( $peak_entry[$dimension*7 - 6] eq "{?}" or $peak_entry[$dimension*7 - 6] eq "null" ) {
                        $crosspeak{$xpkno}{'ass'}[$dimension] = $peak_entry[$dimension*7 - 6];
                    } else {
                        $crosspeak{$xpkno}{'ass'}[$dimension] = $sequence[$number]."_".$number.":".uc($atom);
                    }
                }
                $crosspeak{$xpkno}{'volume'} = $peak_entry[$number_dimensions*7 + 1];
                ++$num_xpk;
            }

			last SWITCH;
		}
		if ( $data_type eq "tsriNMRView" ) {
			while ( $#peak_entry != $number_dimensions*6 + 3 ) {
				if ($new_xpk) { $new_xpk_header .= $line; }	# read through header
				$line = <XPK>;					# storing lines in string
				$line =~ s/^\s+//;				# and splitting the next line
				@peak_entry = split(/\s+/,$line);
			}
			$xpkno = $peak_entry[0];
			if ($new_xpk or $unass_xpk) { $xpk_data{$xpkno} = [ @peak_entry ]; }
			if ($peak_entry[$number_dimensions*6 + 3] != -1) {	# make sure peak has not been marked for deletion
				foreach $dimension (1..$number_dimensions) {
					$crosspeak{$xpkno}{'shift'}[$dimension] = $peak_entry[$dimension*6 - 4];
					($number, $atom) = split('\.', $peak_entry[$dimension*6 - 5]);
					if ( $peak_entry[$dimension*6 - 5] eq "?" or $peak_entry[$dimension*6 - 5] eq "null" ) {
						$crosspeak{$xpkno}{'ass'}[$dimension] = $peak_entry[$dimension*6 - 5];
					} else {
						$crosspeak{$xpkno}{'ass'}[$dimension] = $sequence[$number]."_".$number.":".uc($atom);
					}
				}
				$crosspeak{$xpkno}{'volume'} = $peak_entry[$number_dimensions*6 + 1];
				++$num_xpk;
			}
			last SWITCH;
		}
		if ( $data_type eq "Felix" ) {
			while ( $#peak_entry != $number_dimensions*4 + 1 ) {
				if ($new_xpk) { $new_xpk_header .= $line; }
				$line = <XPK>;
				$line =~ s/^\s+//;
				@peak_entry = split(/\s+/,$line);
			}
			$xpkno = $peak_entry[0];
			if ($new_xpk or $unass_xpk) { $xpk_data{$xpkno} = [ @peak_entry ]; }
			foreach $dimension (1..$number_dimensions) {
				$crosspeak{$xpkno}{'shift'}[$dimension] = $reference_ppm[$dimension] - ($peak_entry[$dimension*4 - 3] - $reference_point[$dimension])*$factor[$dimension];
				$crosspeak{$xpkno}{'ass'}[$dimension] = $peak_entry[$dimension*4];
			}
			
			$vol_data = <VOL>;
			@vol_entry = split(/\s+/,$vol_data);
			if ( $vol_entry[1] == $xpkno ) { $crosspeak{$xpkno}{'volume'} = $vol_entry[2]; }
			else { die "FATAL ERROR! Mismatch in peak and volume lists at crosspeak #$xpkno\n"; }
			++$num_xpk;
			last SWITCH;
		}
	}
}
close (XPK);
print "Read $num_xpk crosspeaks from $xpk_file\n";
print OUT "Read $num_xpk crosspeaks from $xpk_file\n";
if ( $data_type eq "Felix" ) {
	print "Read $num_xpk volumes from $vol_file\n";
	print OUT "Read $num_xpk volumes from $vol_file\n";
	close (VOL);
}
}


sub CHECK_WATER
# Checks that chemical shifts in all dimensions are further
# than water_tolerance from the water peak.  If not, peak is 
# flagged for ignoring and warnings are printed.
{
foreach $dimension (1..$number_dimensions) {
	if ( abs($crosspeak{$xpkno}{'shift'}[$dimension] - $water_shift) <= $water_tolerance) {
		$ignoring = "true";
		$message .= "Crosspeak is too close to water resonance\n";
	}
}
}

sub DETERMINE_BIN
# Determines bin from the volume.  Can set bin boundaries in 
# terms of distances or volumes.  Peaks with negative 
# volumes are flagged for ignoring.  
{
$Rij = "";
if ($crosspeak{$xpkno}{'volume'} <= 0) {
	if ($using_negative_volumes) {
		$crosspeak{$xpkno}{'volume'} *= -1;
	} else {
		$Rij = "???";
		$ignoring = "true";
   		$message .= "Crosspeak has negative volume\n";
		return;
	}
} 
if ($bin_volumes) { 
	$Rij0 = $crosspeak{$xpkno}{'volume'} / $scaling_constant; 
#	$Rij0 = int($Rij0);
	BOUND: {
		if ($Rij0 >= $Bound1) { $Rij = $Bin1; last BOUND; } 
		if ( ($Rij0 >= $Bound2) && ($Rij0 < $Bound1) ) { $Rij = $Bin2; last BOUND; } 
		if ( ($Rij0 >= $Bound3) && ($Rij0 < $Bound2) ) { $Rij = $Bin3; last BOUND; } 
		if ( ($Rij0 >= $Bound4) && ($Rij0 < $Bound3) ) { $Rij = $Bin4; last BOUND; }
		if ( ($Rij0 >= $Bound5) && ($Rij0 < $Bound4) ) { $Rij = $Bin5; last BOUND; }
		if ( ($Rij0 >= $Bound6) && ($Rij0 < $Bound5) ) { $Rij = $Bin6; last BOUND; }
		if ( ($Rij0 >= $Bound7) && ($Rij0 < $Bound6) ) { $Rij = $Bin7; last BOUND; }
		if ( ($Rij0 >= $Bound8) && ($Rij0 < $Bound7) ) { $Rij = $Bin8; last BOUND; }
		if ( ($Rij0 >= $Bound9) && ($Rij0 < $Bound8) ) { $Rij = $Bin9; last BOUND; }
		if ( ($Rij0 >= $Bound10) && ($Rij0 < $Bound9) ) { $Rij = $Bin10; last BOUND; }
	}
	if ($Rij eq "") { 
		$Rij = "???";
		$ignoring = "true";
		$message .= "Crosspeak volume too small.  Manipulated volume: $Rij0\n";
	}
} elsif ($bin_distances) { 
	$crosspeak{$xpkno}{'volume'} = $crosspeak{$xpkno}{'volume'} / $scaling_constant;
	$Rij0 = $intercept + $gradient/(exp((log $crosspeak{$xpkno}{'volume'})/6)); 
	$Rij0 = int($Rij0);
	BOUND: {
		if ($Rij0 <= $Bound1) { $Rij = $Bin1; last BOUND; } 
		if ( ($Rij0 > $Bound1) && ($Rij0 <= $Bound2) ) { $Rij = $Bin2; last BOUND; } 
		if ( ($Rij0 > $Bound2) && ($Rij0 <= $Bound3) ) { $Rij = $Bin3; last BOUND; } 
		if ( ($Rij0 > $Bound3) && ($Rij0 <= $Bound4) ) { $Rij = $Bin4; last BOUND; }
		if ( ($Rij0 > $Bound4) && ($Rij0 <= $Bound5) ) { $Rij = $Bin5; last BOUND; }
		if ( ($Rij0 > $Bound5) && ($Rij0 <= $Bound6) ) { $Rij = $Bin6; last BOUND; }
		if ( ($Rij0 > $Bound6) && ($Rij0 <= $Bound7) ) { $Rij = $Bin7; last BOUND; }
		if ( ($Rij0 > $Bound7) && ($Rij0 <= $Bound8) ) { $Rij = $Bin8; last BOUND; }
		if ( ($Rij0 > $Bound8) && ($Rij0 <= $Bound9) ) { $Rij = $Bin9; last BOUND; }
		if ( ($Rij0 > $Bound9) && ($Rij0 <= $Bound10)) { $Rij = $Bin10;last BOUND; }
	}
	if ($Rij eq "") { 
		$Rij = "???";
		$ignoring = "true";
		$message .= sprintf "Crosspeak volume too small.  Calculated distance: %5.2f A\n", $Rij0;
	}
}
&ADJUST_BIN;
}

sub ADJUST_BIN
# Checks for crosspeak number in adjust list and adjusts 
# to next lowest bin if found. If crosspeak number appears 
# more than once bin will undergo multiple adjustments.
{
foreach $adjust (@adjust_list) { 
	if ($xpkno == $adjust) { 
		print OUT "WARNING: Bin adjusted\n";
		BIN: {
			if ($Rij == $Bin1) { $Rij = $Bin2; last BIN; } 
			if ($Rij == $Bin2) { $Rij = $Bin3; last BIN; }
			if ($Rij == $Bin3) { $Rij = $Bin4; last BIN; }
			if ($Rij == $Bin4) { $Rij = $Bin5; last BIN; }
			if ($Rij == $Bin5) { $Rij = $Bin6; last BIN; }
			if ($Rij == $Bin6) { $Rij = $Bin7; last BIN; }
			if ($Rij == $Bin7) { $Rij = $Bin8; last BIN; }
			if ($Rij == $Bin8) { $Rij = $Bin9; last BIN; }
			if ($Rij == $Bin9) { $Rij = $Bin10;last BIN; }
			if ($Rij == $Bin10) { $Rij = "";   last BIN; }
		}
		if ($Rij eq "") {
			$Rij = "???";
			$ignoring = "true";
			$message .= "Attempted to adjust to lower than smallest bin\n";
		}
	}
}
}


sub FIND_MIRRORS
# comments here
{
$mirrored_shifts = "";
$mirrored_ass = "";
foreach $peak_number ( sort numerically keys %crosspeak ) {
	if ($peak_number != $xpkno) {
		if ($expt_flag < 30 or $expt_flag >= 40) {			# not 3D HSQC-NOESY-HSQC
			if ( $tolerance[$protonA_dim] > $tolerance[$protonB_dim] ) {
				$biggest_tolerance = $tolerance[$protonA_dim];
			} else {
				$biggest_tolerance = $tolerance[$protonB_dim];
			}
			if ( abs($crosspeak{$xpkno}{'shift'}[$protonA_dim] - $crosspeak{$peak_number}{'shift'}[$protonB_dim]) <= $biggest_tolerance and
			     abs($crosspeak{$xpkno}{'shift'}[$protonB_dim] - $crosspeak{$peak_number}{'shift'}[$protonA_dim]) <= $biggest_tolerance ) {
				$mirrored_shifts = "$mirrored_shifts$peak_number,";
			}
			if ( $crosspeak{$xpkno}{'ass'}[$protonA_dim] eq $crosspeak{$peak_number}{'ass'}[$protonB_dim] and
			     $crosspeak{$xpkno}{'ass'}[$protonB_dim] eq $crosspeak{$peak_number}{'ass'}[$protonA_dim] and
			     $crosspeak{$xpkno}{'ass'}[$protonA_dim] ne "{?}" and $crosspeak{$xpkno}{'ass'}[$protonB_dim] ne "{?}" and
			     $crosspeak{$xpkno}{'ass'}[$protonA_dim] ne "?" and $crosspeak{$xpkno}{'ass'}[$protonB_dim] ne "?" and
			     $crosspeak{$xpkno}{'ass'}[$protonA_dim] ne "null" and $crosspeak{$xpkno}{'ass'}[$protonB_dim] ne "null" ) {
				$mirrored_ass = "$mirrored_ass$peak_number,";
			}
		} else {
			if ( $tolerance[$heteroA_dim] > $tolerance[$heteroB_dim] ) {
				$biggest_tolerance = $tolerance[$heteroA_dim];
			} else {
				$biggest_tolerance = $tolerance[$heteroB_dim];
			}
			if ( abs($crosspeak{$xpkno}{'shift'}[$heteroA_dim] - $crosspeak{$peak_number}{'shift'}[$heteroB_dim]) <= $biggest_tolerance and
			     abs($crosspeak{$xpkno}{'shift'}[$heteroB_dim] - $crosspeak{$peak_number}{'shift'}[$heteroA_dim]) <= $biggest_tolerance ) {
				$mirrored_shifts = "$mirrors$peak_number,";
			}
		}
		
	}
}
if ($mirrored_shifts =~ /,/) {
	chop($mirrored_shifts);
	print OUT "Possible mirror peaks are $mirrored_shifts\n";
} else {
	print OUT "No possible mirror peaks found\n";
}
if ($mirrored_ass =~ /,/) {
	chop($mirrored_ass);
	print OUT "Peaks assigned as mirrors are $mirrored_ass\n";
} else {
	print OUT "No peaks assigned as mirrors\n";
}
}

sub FIND_POSSIBILITIES
# Creates initial array 'temp' with list of possible assignments 
# for each dimension. The name and deviation from assigned 
# chemical shift are stored.  Unassigned atoms whose mean BMRB 
# are within a user specified number of standard deviations are 
# also included as possibilities.  In these cases the chemical 
# shift deviation is set to '????'.
{
@temp = ();
foreach $dimension (1..$number_dimensions) {
	foreach $name (keys %{$assignment[$dimension]}) {
		$test_shift = $assignment[$dimension]{$name};
		($residue, $atom) = split(':', $name);
		SWITCH: {
			if ($test_shift ne "unassigned") { 		# look for assigned atoms first
				if ( abs($crosspeak{$xpkno}{'shift'}[$dimension] - $test_shift) <= $tolerance[$dimension] ) {
					push @{$temp[$dimension]}, $name;
					push @{$temp[$dimension + $number_dimensions]}, $crosspeak{$xpkno}{'shift'}[$dimension] - $test_shift;
					last SWITCH;
				}					# if near lower edge check upper edge as well
				if ( $crosspeak{$xpkno}{'shift'}[$dimension] + $tolerance[$dimension] > $lower[$dimension] and
				    abs($crosspeak{$xpkno}{'shift'}[$dimension] - $swp[$dimension] - $test_shift) <= $tolerance[$dimension] ) {
					push @{$temp[$dimension]}, $name;
					push @{$temp[$dimension + $number_dimensions]}, $crosspeak{$xpkno}{'shift'}[$dimension] - $swp[$dimension] - $test_shift;
					last SWITCH;
				}					# if near upper edge check lower edge as well
				if ( $crosspeak{$xpkno}{'shift'}[$dimension] - $tolerance[$dimension] < $upper[$dimension] and 
				    abs($crosspeak{$xpkno}{'shift'}[$dimension] + $swp[$dimension] - $test_shift) <= $tolerance[$dimension] ) {
					push @{$temp[$dimension]}, $name;
					push @{$temp[$dimension + $number_dimensions]}, $crosspeak{$xpkno}{'shift'}[$dimension] + $swp[$dimension] - $test_shift;
					last SWITCH;
				}
			} else {					# add unassigned nuclei as possibilities
				($residue_name, $residue_number) = split('_', $residue);
				if ( abs( $crosspeak{$xpkno}{'shift'}[$dimension] - $bmrb{$residue_name}{$atom}[$dimension]{"mean"} ) <= $bmrb{$residue_name}{$atom}[0]{"std dev"} * $number_BMRB_stddevs ) {
					push @{$temp[$dimension]}, $name;
					push @{$temp[$dimension + $number_dimensions]}, "????";
					last SWITCH;
				}
				if ( $crosspeak{$xpkno}{'shift'}[$dimension] + $tolerance[$dimension] > $lower[$dimension] and
				    abs($crosspeak{$xpkno}{'shift'}[$dimension] - $swp[$dimension] - $bmrb{$residue_name}{$atom}[$dimension]{"mean"}) <= $tolerance[$dimension] ) {
					push @{$temp[$dimension]}, $name;
					push @{$temp[$dimension + $number_dimensions]}, "????";
					last SWITCH;
				}
				if ( $crosspeak{$xpkno}{'shift'}[$dimension] - $tolerance[$dimension] < $upper[$dimension] and 
				    abs($crosspeak{$xpkno}{'shift'}[$dimension] + $swp[$dimension] - $bmrb{$residue_name}{$atom}[$dimension]{"mean"}) <= $tolerance[$dimension] ) {
					push @{$temp[$dimension]}, $name;
					push @{$temp[$dimension + $number_dimensions]}, "????";
					last SWITCH;
				}
			}
		}
	}
}

# Once all possibilities in each dimension are found then every
# possible combination is checked to ensure that it is a possible 
# NOE for this type of experiment. Acceptable NOEs are pushed onto 
# a new array 'possibility'. Peaks with a diagonal or geminal 
# possibility are flagged for ignoring.

@possibility = ();
$diagonal_or_geminal = "";
EXPERIMENT_TYPE: {
	if ($expt_flag >= 0 and $expt_flag < 10) {				# 2D NOESY
		foreach $i (0..$#{$temp[$protonA_dim]}) {
			foreach $j (0..$#{$temp[$protonB_dim]}) {
				($protonA_residue, $protonA_atom) = split(":", $temp[$protonA_dim][$i]);
				($protonB_residue, $protonB_atom) = split(":", $temp[$protonB_dim][$j]);
				$possible_NOE = "";
				if ( $temp[$protonA_dim + $number_dimensions][$i] ne "????" or 
				     $temp[$protonB_dim + $number_dimensions][$j] ne "????" ) {
					$possible_NOE = "true";			# one or both sides of the possibility 
				} else {					# is an assigned atom
					if ( $temp[$protonA_dim][$i] eq $temp[$protonB_dim][$j] and 
					     abs($crosspeak{$xpkno}{'shift'}[$protonA_dim] - $crosspeak{$xpkno}{'shift'}[$protonB_dim]) <= 2 * $tolerance[$protonA_dim] ) {
						$possible_NOE = "true";		# possibility includes two unassigned atoms
					}					# and the difference in chemical shift
				}						# is within the specified tolerance
				if ( $possible_NOE ) {
					$info[$protonA_dim] = $temp[$protonA_dim][$i];
					$info[$protonB_dim] = $temp[$protonB_dim][$j];
					$info[$protonA_dev] = $temp[$protonA_dev][$i];
					$info[$protonB_dev] = $temp[$protonB_dev][$j];
					if ( $protonA_residue eq $protonB_residue and 
					     substr($protonA_atom,0,2) eq substr($protonB_atom,0,2) ) {
						$diagonal_or_geminal = "true";
						$info[$diageminal] = "true";
	 				} else {
						$info[$diageminal] = "";
					}
					push @possibility, [ @info ];
				}
			}
		}
		last EXPERIMENT_TYPE;
	}
	if ($expt_flag >= 10 and $expt_flag < 30) {			# 3D 15N NOESY and 3D 13C NOESY
		foreach $i (0..$#{$temp[$protonA_dim]}) {
			foreach $j (0..$#{$temp[$heteroB_dim]}) {
				foreach $k (0..$#{$temp[$protonB_dim]}) {
					($protonA_residue, $protonA_atom) = split(":", $temp[$protonA_dim][$i]);
					($heteroB_residue, $heteroB_atom) = split(":", $temp[$heteroB_dim][$j]);
					($protonB_residue, $protonB_atom) = split(":", $temp[$protonB_dim][$k]);
					if ($protonB_residue eq $heteroB_residue) {
						if (length($heteroB_atom) > 1) {
							$heteroB_pos = substr($heteroB_atom,1);
						} else { 
							if ($heteroB_atom eq "N") { $heteroB_pos = "N"; }
							elsif ($heteroB_atom eq "C") { $heteroB_pos = "O"; }
						}
						if (length($protonB_atom) > 1) {
							$protonB_pos = substr($protonB_atom,1,length($heteroB_pos));
						} else {
							if ($protonB_atom eq "H") { $protonB_pos = "N"; }
						}
						$possible_NOE = "";
						if ($protonB_pos eq $heteroB_pos) {
							if ( $temp[$protonA_dim + $number_dimensions][$i] ne "????" or 
							     $temp[$protonB_dim + $number_dimensions][$k] ne "????" ) {
								$possible_NOE = "true";	
							} else {
								if ( $temp[$protonA_dim][$i] eq $temp[$protonB_dim][$k] and 
								     abs($crosspeak{$xpkno}{'shift'}[$protonA_dim] - $crosspeak{$xpkno}{'shift'}[$protonB_dim]) <= 2 * $tolerance[$protonA_dim] ) {
									$possible_NOE = "true";
								}
							}
						}
						if ( $possible_NOE ) {
							$info[$protonA_dim] = $temp[$protonA_dim][$i];
							$info[$heteroB_dim] = $temp[$heteroB_dim][$j];
							$info[$protonB_dim] = $temp[$protonB_dim][$k];
							$info[$protonA_dev] = $temp[$protonA_dev][$i];
							$info[$heteroB_dev] = $temp[$heteroB_dev][$j];
							$info[$protonB_dev] = $temp[$protonB_dev][$k];
							if ( $protonA_residue eq $protonB_residue and 
							     substr($protonA_atom,0,2) eq substr($protonB_atom,0,2) ) {
								$diagonal_or_geminal = "true";
								$info[$diageminal] = "true";
	 						} else {
								$info[$diageminal] = "";
							}
							push @possibility, [ @info ];
						}
					}
				}
			}
		}
		last EXPERIMENT_TYPE;
	}
	if ($expt_flag >= 30 and $expt_flag < 40) {			# 3D HSQC-NOESY-HSQC
		$protonA_dim = 10;					# Store protonA_dim name at end of info 
		foreach $i (0..$#{$temp[$heteroA_dim]}) {		# list. Need for measuring distances and
			foreach $j (0..$#{$temp[$heteroB_dim]}) {	# writing restraints.
				foreach $k (0..$#{$temp[$protonB_dim]}) {
					($heteroA_residue, $heteroA_atom) = split(":", $temp[$heteroA_dim][$i]);
					($heteroB_residue, $heteroB_atom) = split(":", $temp[$heteroB_dim][$j]);
					($protonB_residue, $protonB_atom) = split(":", $temp[$protonB_dim][$k]);
					if ($protonB_residue eq $heteroB_residue) {
						if (length($heteroB_atom) > 1) {
							$heteroB_pos = substr($heteroB_atom,1);
						} else { 
							if ($heteroB_atom eq "N") { $heteroB_pos = "N"; }
							elsif ($heteroB_atom eq "C") { $heteroB_pos = "O"; }
						}
						if (length($protonB_atom) > 1) {
							$protonB_pos = substr($protonB_atom,1,length($heteroB_pos));
						} else {
							if ($protonB_atom eq "H") { $protonB_pos = "N"; }
						}
						$possible_NOE = "";
						if ($protonB_pos eq $heteroB_pos) {
							if ( $temp[$heteroA_dim + $number_dimensions][$i] ne "????" or 
							     $temp[$heteroB_dim + $number_dimensions][$j] ne "????" ) {
								$possible_NOE = "true";	
							} else {
								if ( $temp[$heteroA_dim][$i] eq $temp[$heteroB_dim][$j] and 
								     abs($crosspeak{$xpkno}{'shift'}[$heteroA_dim] - $crosspeak{$xpkno}{'shift'}[$heteroB_dim]) <= 2 * $tolerance[$heteroA_dim] ) {
									$possible_NOE = "true";
								}
							}
						}
						if ( $possible_NOE ) {
							$info[$heteroA_dim] = $temp[$heteroA_dim][$i];
							$info[$heteroB_dim] = $temp[$heteroB_dim][$j];
							$info[$protonB_dim] = $temp[$protonB_dim][$k];
							$info[$heteroA_dev] = $temp[$heteroA_dev][$i];
							$info[$heteroB_dev] = $temp[$heteroB_dev][$j];
							$info[$protonB_dev] = $temp[$protonB_dev][$k];
							NITROGEN_ATTACHED_PROTON: {
								if ($heteroA_atom eq "NN" or $heteroA_atom eq "N") {	# backbone amide
									$info[$protonA_dim] = "$heteroA_residue:HN";
									last NITROGEN_ATTACHED_PROTON;
								}
								if ($heteroA_atom eq "ND1") {	# histidine
									$info[$protonA_dim] = "$heteroA_residue:HD1";
									last NITROGEN_ATTACHED_PROTON;
								}
								if ($heteroA_atom eq "ND2") {	# asparagine
									$info[$protonA_dim] = "$heteroA_residue:QD2";
									last NITROGEN_ATTACHED_PROTON;
								}
								if ($heteroA_atom eq "NE") {	# arginine
									$info[$protonA_dim] = "$heteroA_residue:HE";
									last NITROGEN_ATTACHED_PROTON;
								}
								if ($heteroA_atom eq "NE1") {	# tryptophan
									$info[$protonA_dim] = "$heteroA_residue:HE1";
									last NITROGEN_ATTACHED_PROTON;
								}
								if ($heteroA_atom eq "NE2" and substr($heteroA_residue,0,3) eq "HIS") {
									$info[$protonA_dim] = "$heteroA_residue:HE2";
									last NITROGEN_ATTACHED_PROTON;
								}
								if ($heteroA_atom eq "NE2" and substr($heteroA_residue,0,3) eq "GLN") {
									$info[$protonA_dim] = "$heteroA_residue:QE2";
									last NITROGEN_ATTACHED_PROTON;
								}
								if ($heteroA_atom eq "NZ") {	# lysine
									$info[$protonA_dim] = "$heteroA_residue:QNZ";
									last NITROGEN_ATTACHED_PROTON;
								}
							}
							if ( $heteroA_residue eq $heteroB_residue and $heteroA_atom eq $heteroB_atom ) {
								$diagonal_or_geminal = "true";
								$info[$diageminal] = "true";
	 						} else {
								$info[$diageminal] = "";
							}
							push @possibility, [ @info ];
						}
					}
				}
			}
		}
		last EXPERIMENT_TYPE;
	}
	if ($expt_flag >= 40 and $expt_flag < 50) {			# 4D experiments
		foreach $i (0..$#{$temp[$protonA_dim]}) {
			foreach $j (0..$#{$temp[$heteroA_dim]}) {
				foreach $k (0..$#{$temp[$heteroB_dim]}) {
					foreach $l (0..$#{$temp[$protonB_dim]}) {
						($protonA_residue, $protonA_atom) = split(":", $temp[$protonA_dim][$i]);
						($heteroA_residue, $heteroA_atom) = split(":", $temp[$heteroA_dim][$j]);
						($heteroB_residue, $heteroB_atom) = split(":", $temp[$heteroB_dim][$k]);
						($protonB_residue, $protonB_atom) = split(":", $temp[$protonB_dim][$l]);
						if ($protonA_residue eq $heteroA_residue and $protonB_residue eq $heteroB_residue) {
							if (length($heteroA_atom) > 1) {
								$heteroA_pos = substr($heteroA_atom,1);
							} else { 
								if ($heteroA_atom eq "N") { $heteroA_pos = "N"; }
								elsif ($heteroA_atom eq "C") { $heteroA_pos = "O"; }
							}
							if (length($protonA_atom) > 1) {
								$protonA_pos = substr($protonA_atom,1,length($heteroA_pos));
							} else {
								if ($protonA_atom eq "H") { $protonA_pos = "N"; }
							}
							if (length($heteroB_atom) > 1) {
								$heteroB_pos = substr($heteroB_atom,1);
							} else { 
								if ($heteroB_atom eq "N") { $heteroB_pos = "N"; }
								elsif ($heteroB_atom eq "C") { $heteroB_pos = "O"; }
							}
							if (length($protonB_atom) > 1) {
								$protonB_pos = substr($protonB_atom,1,length($heteroB_pos));
							} else {
								if ($protonB_atom eq "H") { $protonB_pos = "N"; }
							}
							$possible_NOE = "";
							if ($protonA_pos eq $heteroA_pos and $protonB_pos eq $heteroB_pos ) {
								if ( $temp[$protonA_dim + $number_dimensions][$i] ne "????" or 
								     $temp[$protonB_dim + $number_dimensions][$l] ne "????" ) {
									$possible_NOE = "true";	
								} else {
									if ( $temp[$protonA_dim][$i] eq $temp[$protonB_dim][$l] and 
									     abs($crosspeak{$xpkno}{'shift'}[$protonA_dim] - $crosspeak{$xpkno}{'shift'}[$protonB_dim]) <= 2 * $tolerance[$protonA_dim] ) {
										$possible_NOE = "true";
									}
								}
							}
							if ( $possible_NOE ) {
								$info[$protonA_dim] = $temp[$protonA_dim][$i];
								$info[$heteroA_dim] = $temp[$heteroA_dim][$j];
								$info[$heteroB_dim] = $temp[$heteroB_dim][$k];
								$info[$protonB_dim] = $temp[$protonB_dim][$l];
								$info[$protonA_dev] = $temp[$protonA_dev][$i];
								$info[$heteroA_dev] = $temp[$heteroA_dev][$j];
								$info[$heteroB_dev] = $temp[$heteroB_dev][$k];
								$info[$protonB_dev] = $temp[$protonB_dev][$l];
								if ( $protonA_residue eq $protonB_residue and substr($protonA_atom,0,2) eq substr($protonB_atom,0,2) ) {
									$diagonal_or_geminal = "true";				
									$info[$diageminal] = "true";
	 							} else {
									$info[$diageminal] = "";
								}
								push @possibility, [ @info ];
							}
	 					}
					}
				}
			}
		}
		last EXPERIMENT_TYPE;
	}
	if ($expt_flag >= 50 and $expt_flag < 60) {			# select filter experiments?
		die "Haven't implemented select filter experiments yet\n";
		last EXPERIMENT_TYPE;
	}
	if ($expt_flag >= 60 and $expt_flag < 70) {			# 3D aromatic NOESY
		foreach $i (0..$#{$temp[$protonA_dim]}) {
			foreach $j (0..$#{$temp[$heteroB_dim]}) {
				foreach $k (0..$#{$temp[$protonB_dim]}) {
					($protonA_residue, $protonA_atom) = split(":", $temp[$protonA_dim][$i]);
					($heteroB_residue, $heteroB_atom) = split(":", $temp[$heteroB_dim][$j]);
					($protonB_residue, $protonB_atom) = split(":", $temp[$protonB_dim][$k]);
					if ((substr($protonB_residue,0,3) eq "PHE" and $protonB_residue eq $heteroB_residue) or
					    (substr($protonB_residue,0,3) eq "TYR" and $protonB_residue eq $heteroB_residue) or
					    (substr($protonB_residue,0,3) eq "TRP" and $protonB_residue eq $heteroB_residue) or
					    (substr($protonB_residue,0,3) eq "HIS" and $protonB_residue eq $heteroB_residue)) {
						if (length($heteroB_atom) > 1) {
							$heteroB_pos = substr($heteroB_atom,1);
						} else { 
							if ($heteroB_atom eq "N") { $heteroB_pos = "N"; }
							elsif ($heteroB_atom eq "C") { $heteroB_pos = "O"; }
						}
						if (length($protonB_atom) > 1) {
							$protonB_pos = substr($protonB_atom,1,length($heteroB_pos));
						} else {
							if ($protonB_atom eq "H") { $protonB_pos = "N"; }
						}
						$possible_NOE = "";
						if ($protonB_pos eq $heteroB_pos) {
							if ($temp[$protonA_dim + $number_dimensions][$i] ne "????" or 
							    $temp[$protonB_dim + $number_dimensions][$k] ne "????") {
								$possible_NOE = "true";	
							} else {
								if ($temp[$protonA_dim][$i] eq $temp[$protonB_dim][$k] and 
								    abs($crosspeak{$xpkno}{'shift'}[$protonA_dim] - $crosspeak{$xpkno}{'shift'}[$protonB_dim]) <= 2 * $tolerance[$protonA_dim] ) {
									$possible_NOE = "true";
								}
							}
						}
						if ( $possible_NOE ) {
							$info[$protonA_dim] = $temp[$protonA_dim][$i];
							$info[$heteroB_dim] = $temp[$heteroB_dim][$j];
							$info[$protonB_dim] = $temp[$protonB_dim][$k];
							$info[$protonA_dev] = $temp[$protonA_dev][$i];
							$info[$heteroB_dev] = $temp[$heteroB_dev][$j];
							$info[$protonB_dev] = $temp[$protonB_dev][$k];
							if ( $protonA_residue eq $protonB_residue and 
							     substr($protonA_atom,0,2) eq substr($protonB_atom,0,2) ) {
								$diagonal_or_geminal = "true";
								$info[$diageminal] = "true";
	 						} else {
								$info[$diageminal] = "";
							}
							push @possibility, [ @info ];
						}
					}
				}
			}
		}
		last EXPERIMENT_TYPE;
	}
}
$number_possibilities = $#possibility + 1;
if ($number_possibilities == 0) {
	$message .= "No possible assignments found\n";
} elsif ($number_possibilities == 1) {
	print OUT "1 possibility found\n";
} elsif ($number_possibilities > 1) {
	print OUT "$number_possibilities possibilities found\n";
}
&WRITE_INFO;
}

sub FILTER_ASSIGNMENTS
# Filters the possibilities by the existing assignments.
# Dimensions in which the assignment is 'null' or '?'
# are treated as unassigned. If none of the possibilities
# match the existing assignments then the assignments are
# ignored and a warning is printed.
{
@temp = ();
$number_before = $number_possibilities;
$diagonal_or_geminal = "";
foreach $possibility_number (0..$#possibility) {
	$assignments_match = "true";
	foreach $dimension (1..$number_dimensions) {
		if ($crosspeak{$xpkno}{'ass'}[$dimension] ne "null" and 
		    $crosspeak{$xpkno}{'ass'}[$dimension] ne "?" and 
		    $crosspeak{$xpkno}{'ass'}[$dimension] ne "{?}" and 
		    $crosspeak{$xpkno}{'ass'}[$dimension] ne $possibility[$possibility_number][$dimension]) {
			$assignments_match = "";
		}
	}
	if ($assignments_match) {
		push @temp, $possibility[$possibility_number];
		if ($possibility[$possibility_number][$diageminal] eq "true") { $diagonal_or_geminal = "true"; }
	}
}
if ($#temp == -1) {
	print OUT "Assignments don't match possibilities. Ignoring assignments\n";
} else { 
	@possibility = @temp;
}
$number_possibilities = $#possibility + 1;
$number_eliminated = $number_before - $number_possibilities;
if ($number_possibilities == 0) {
	$message .= "No possible assignment after assignment filtering\n";
} elsif ($number_eliminated == 1) {
	print OUT "Assignment filtering eliminated 1 possibility\n";
	&WRITE_INFO;
} elsif ($number_eliminated > 1) {
	print OUT "Assignment filtering eliminated $number_eliminated possibilities\n";
	&WRITE_INFO;
}
}

sub MEASURE_DISTANCES
# Each possibility found is expanded according to
# the pseudoatoms listed in the map file.  The distance
# between all combinations of expanded atoms is 
# measured in all structures in the ensemble.
# The shortest distance is found, and the average
# distance is calculated.  
{
$shortest_average = 1000;
foreach $possibility_number (0..$#possibility) {
	$shortest_distance = 1000;
	$sum_distances = 0;
	$number_distances = 0;
	($resname_A, $resnum_A, $atom_A) = ($possibility[$possibility_number][$protonA_dim] =~ /([A-Z]{3,4})_(\d+):(\S+)/);
	($resname_B, $resnum_B, $atom_B) = ($possibility[$possibility_number][$protonB_dim] =~ /([A-Z]{3,4})_(\d+):(\S+)/);
	foreach $mapNameA (@{$map{$resname_A}{$atom_A}}) {
		$poss_A = $resname_A."_".$resnum_A.":".$mapNameA;
		foreach $mapNameB (@{$map{$resname_B}{$atom_B}}) {
			$poss_B = $resname_B."_".$resnum_B.":".$mapNameB;
			foreach $structure_number (0..$#atom) {
				$distance = sqrt( 
		   ($atom[$structure_number]{$poss_A}{'xpos'} - $atom[$structure_number]{$poss_B}{'xpos'})**2 + 
		   ($atom[$structure_number]{$poss_A}{'ypos'} - $atom[$structure_number]{$poss_B}{'ypos'})**2 + 
		   ($atom[$structure_number]{$poss_A}{'zpos'} - $atom[$structure_number]{$poss_B}{'zpos'})**2
		   );
				$sum_distances = $sum_distances + $distance;
				$number_distances++;
				if ($distance < $shortest_distance) { $shortest_distance = $distance; }
			}
		}
	}
	if ($number_distances == 0) { die "$map_file seems to be missing an atom name.\nEither $resname_A $atom_A or $resname_B $atom_B.\n"; }
	$average_distance = $sum_distances/$number_distances;
	if ($average_distance < $shortest_average) { $shortest_average = $average_distance; }
	$possibility[$possibility_number][$shortest] = $shortest_distance;
	$possibility[$possibility_number][$average] = $average_distance;
}
}

sub WRITE_INFO
# Prints a list of the possibilities and their 
# chemical shift deviations, shortest and average 
# distances in the ensemble, and NOE contributions.
{
foreach $possibility_number (0..$#possibility) {
	foreach $i (1..$number_dimensions) {
		printf OUT "%-13s", $possibility[$possibility_number][$i];
	}
	foreach $i (1..$number_dimensions) {
		if ($possibility[$possibility_number][$i] eq "????") { print OUT "  ???? "; }
		else { printf OUT "%7.3f", $possibility[$possibility_number][$i+$number_dimensions]; }
	}
	if ($possibility[$possibility_number][$average]) {
		foreach $i ($shortest, $average) {
			printf OUT "%7.2f", $possibility[$possibility_number][$i];
		}
	}
	if ($possibility[$possibility_number][0]) { printf OUT "  %7.5f", $possibility[$possibility_number][0]; }
	print OUT "\n";
}
}

sub FILTER_DISTANCES
# Ensures that every possibility has an average
# distance in the ensemble less than the specified
# distance cutoff.
{
@contributions = ();
@temp = ();
$number_before = $#possibility + 1;
foreach $possibility_number (0..$#possibility) {
	if ( $possibility[$possibility_number][$average] <= $distance_cutoff ) { 
		push @temp, $possibility[$possibility_number];
		push @contributions, $possibility[$possibility_number][0];
	}
}
@possibility = @temp;
$number_possibilities = $#possibility + 1;
$number_eliminated = $number_before - $number_possibilities;
if ($number_possibilities == 0) {
	$message .= "No possible assignment after distance filtering\n";
} elsif ($number_eliminated == 1) {
	print OUT "Distance filtering eliminated 1 possibility\n";
	&WRITE_INFO;
} elsif ($number_eliminated > 1) {
	print OUT "Distance filtering eliminated $number_eliminated possibilities\n";
	&WRITE_INFO;
}
}

sub FILTER_SECONDARY
# comments
{
@temp = ();
$number_before = $#possibility + 1;
foreach $possibility_number (0..$#possibility) {
	$discarding = "";
	$secondary_i = "i";	# these two variables have to have different values
	$secondary_j = "j";	# therefore they can't be undefined
	($i, $atom_i) = split(":", $possibility[$possibility_number][$protonA_dim]);
	($j, $atom_j) = split(":", $possibility[$possibility_number][$protonB_dim]);
	($res_i, $num_i) = split("_", $i);
	($res_j, $num_j) = split("_", $j);

	# make sure i comes before j
	if ($num_j < $num_i) {
		($num_i, $num_j) = ($num_j, $num_i);
		($atom_i, $atom_j) = ($atom_j, $atom_i);
	}

	# identify the secondary structure for i and j
	foreach $element ( keys %sse) {
		if ($num_i >= $sse{$element}{'start'} && $num_i <= $sse{$element}{'end'} ) {
			$secondary_i = $element;
		}
		if ($num_j >= $sse{$element}{'start'} && $num_j <= $sse{$element}{'end'} ) {
			$secondary_j = $element;
		}
	}

	# build up restraint name
	$atom_i = position($atom_i);
	$atom_j = position($atom_j);
	$diff = $num_j - $num_i;
	$restraint = "$atom_i$atom_j$diff,";

	# decide whether or not to discard
	if ($secondary_i eq $secondary_j) {
		if ($secondary_i =~ /helix/ and $exclude{'helix'} =~ $restraint ) { $discarding = "true"; }
		if ($secondary_i =~ /strand/ and $exclude{'strand'} =~ $restraint ) { $discarding = "true"; }
	}
	if ($exclude{'all'} =~ /$restraint/) { $discarding = "true"; }
	push @temp, $possibility[$possibility_number] unless $discarding;
}
@possibility = @temp;
$number_possibilities = $#possibility + 1;
$number_eliminated = $number_before - $number_possibilities;
if ($number_possibilities == 0) {
	$message .= "No possible assignment after secondary structure filtering\n";
} elsif ($number_eliminated == 1) {
	print OUT "Secondary structure filtering eliminated 1 possibility\n";
	&WRITE_INFO;
} elsif ($number_eliminated > 1) {
	print OUT "Secondary structure filtering eliminated $number_eliminated possibilities\n";
	&WRITE_INFO;
}
}

sub position {
	my $atom_name = shift(@_);
	if ($atom_name =~ /A/) { $position = "a"; return $position; }
	if ($atom_name =~ /B/) { $position = "b"; return $position; }
	if ($atom_name =~ /G/) { $position = "g"; return $position; }
	if ($atom_name =~ /D/) { $position = "d"; return $position; }
	if ($atom_name =~ /E/) { $position = "e"; return $position; }
	if ($atom_name =~ /N/) { $position = "N"; return $position; }
}

sub CALC_CONTRIBUTIONS
# Calculates the contribution to the total volume 
# of the NOE crosspeak that each possiblity makes.
# This contribution is stored in an element of the 
# array 'possibility' and in another array 'contributions'.
# 'contributions' is then sorted from largest to smallest.
{
@contributions= ();
$sum_contributions = 0;
foreach $possibility_number (0..$#possibility) {
	$sum_contributions += $possibility[$possibility_number][$shortest]**-6;
}
foreach $possibility_number (0..$#possibility) {
	$contribution = $possibility[$possibility_number][$shortest]**-6 / $sum_contributions;
	$possibility[$possibility_number][0] = $contribution;
	push @contributions, $contribution;
}
@contributions = sort largest_first @contributions;
}

sub largest_first { $b <=> $a; }
# Sorting routine for the contribution filter.

sub FILTER_CONTRIBUTIONS
{
$sum = 0;
@temp = ();
$number_before = $#possibility + 1;
CONTRIBUTION:
foreach $i (0..$#contributions) {
	foreach $possibility_number (0..$#possibility) {
		if ($possibility[$possibility_number][0] == $contributions[$i]) {
			push @temp, $possibility[$possibility_number];
			$sum += $possibility[$possibility_number][0];
			splice(@possibility, $possibility_number, 1);
			if ($sum > $contribution_cutoff) { last CONTRIBUTION; }
			next CONTRIBUTION;
		}
	}
}
@possibility = @temp;
$number_possibilities = $#possibility + 1;
$number_eliminated = $number_before - $number_possibilities;
if ($number_possibilities == 0) {
	$message .= "No possible assignment after contribution filtering\n";
} elsif ($number_eliminated == 1) {
	print OUT "Contribution filtering eliminated 1 possibility\n";
	&WRITE_INFO;
} elsif ($number_eliminated > 1) {
	print OUT "Contribution filtering eliminated $number_eliminated possibilities\n";
	&WRITE_INFO;
}
}

sub WRITE_RESTRAINTS
# Don't write restraint if peak is in ignore list,
# if peak is diagonal or geminal, if peak is
# ambiguous and the number of possibilities is 
# greater than the number being accepted, or if
# ignore flag has been set previously.
{
if (grep(/^$xpkno$/,@ignore_list)) { 
	$message .= "Crosspeak in ignore list\n";
	$ignoring = "true";
}
if ($number_possibilities == 0) {
	$ignoring = "true";	
}
if ($diagonal_or_geminal) {
	$message .= "One of the possibilities is diagonal or geminal\n";
	$ignoring = "true";
}
if ($accepted_possibilities != 0 and $number_possibilities > $accepted_possibilities) {
	$message .= "More possibilities ($number_possibilities) than accepting ($accepted_possibilities)\n";
	$ignoring = "true";	
}
$found = "";
foreach $i (0..$#possibility) { if ($possibility[$i][$average] == $shortest_average) { $found = "true"; last; } }
if ($shortest_average and $shortest_average < $distance_cutoff and not $found ) {
	$message .= "Possibility with lowest mean distance eliminated by filters\n";
}
if ( not $ignoring ) {
	$number_ass_A = 0;
	$number_ass_B = 0;
	$ambigs_A = "";
	$ambigs_B = "";
	# make lists of the possible assignments for the A and B
	# proton dimensions.  Count the number of assignemnts
	# in those dimensions as well
	foreach $i (0..$#possibility) {
		($resname_A, $resnum_A, $atom_A) = ($possibility[$i][$protonA_dim] =~ /([A-Z]{3,4})_(\d+):(\S+)/);
		if ($ambigs_A !~ /\s$atom_A\s$resnum_A\s/) {
			$ambigs_A .= " $atom_A $resnum_A ";
			$number_ass_A++;
		}
		($resname_B, $resnum_B, $atom_B) = ($possibility[$i][$protonB_dim] =~ /([A-Z]{3,4})_(\d+):(\S+)/);
		if ($ambigs_B !~ /\s$atom_B\s$resnum_B\s/) {
			$ambigs_B .= " $atom_B $resnum_B ";
			$number_ass_B++;
		}
	}
	# decide how to write restraints based on the number
	# of assignments for each dimension. use the lists
	# generated above in the ambig map. Check the possiblity
	# against the assignment before calling "preassigned"
	print OUT "Restraint written as :\n";
	CHECK: {
		if ($number_ass_A > 1 and $number_ass_B > 1) {	# Double Ambig
			$label_A = "$expt_flag-$protonA_dim:$xpkno";
			$label_B = "$expt_flag-$protonB_dim:$xpkno";
			print OUT "0 AMB $label_A ($ambigs_A) --- 0 AMB $label_B ($ambigs_B)\n";
			printf UPL "   0 AMB  %-10s    0 AMB  %-10s  ", $label_A,$label_B;
			printf UPL "%4.2f   #   %4d:%-2d  Double Ambig\n", $Rij,$xpkno,$expt_flag;
			print AMBIG_MAP "AMBIG $label_A\t=$ambigs_A\n";
			print AMBIG_MAP "AMBIG $label_B\t=$ambigs_B\n";
			last CHECK;
		}
		if ($number_ass_A > 1 and $number_ass_B = 1) {	# Single Ambig (A)
			$label_A = "$expt_flag-$protonA_dim:$xpkno";
			print OUT "0 AMB $label_A ($ambigs_A) --- $possibility[0][$protonB_dim]\n";
			printf UPL "   0 AMB  %-10s  %3d %-4s %-10s  ", $label_A,$resnum_B,$resname_B,$atom_B;
			printf UPL "%4.2f   #   %4d:%-2d  Single Ambig\n", $Rij,$xpkno,$expt_flag;
			print AMBIG_MAP "AMBIG $label_A\t=$ambigs_A\n";
			last CHECK;
		}
		if ($number_ass_A = 1 and $number_ass_B > 1) {	# Single Ambig (B)
			$label_B = "$expt_flag-$protonB_dim:$xpkno";
			print OUT "$possibility[0][$protonA_dim] --- 0 AMB $label_B ($ambigs_B)\n";
			printf UPL " %3d %-4s %-10s    0 AMB  %-10s  ", $resnum_A,$resname_A,$atom_A,$label_B;
			printf UPL "%4.2f   #   %4d:%-2d  Single Ambig\n", $Rij,$xpkno,$expt_flag;
			print AMBIG_MAP "AMBIG $label_B\t=$ambigs_B\n";
			last CHECK;
		}
		if ($using_assignments and $number_ass_A = 1 and $number_ass_B = 1 and
		    $crosspeak{$xpkno}{'ass'}[$protonA_dim] =~ "$resnum_A:$atom_A" and
		    $crosspeak{$xpkno}{'ass'}[$protonB_dim] =~ "$resnum_B:$atom_B") { 	# Preassigned
			printf UPL " %3d %-4s %-10s  ", $resnum_A,$resname_A,$atom_A;
			printf UPL "%3d %-4s %-10s  ", $resnum_B,$resname_B,$atom_B;
			printf UPL "%4.2f   #   %4d:%-2d  Preassigned   %5.2f,%5.2f A\n", $Rij,$xpkno,$expt_flag,$possibility[0][$shortest],$possibility[0][$average];
			print OUT "$possibility[0][$protonA_dim] --- $possibility[0][$protonB_dim]\n";
			last CHECK;
		}
		if ($number_ass_A = 1 and $number_ass_B = 1) {	# Unique
			printf UPL " %3d %-4s %-10s  ", $resnum_A,$resname_A,$atom_A;
			printf UPL "%3d %-4s %-10s  ", $resnum_B,$resname_B,$atom_B;
			printf UPL "%4.2f   #   %4d:%-2d  Unique        %5.2f,%5.2f A\n", $Rij,$xpkno,$expt_flag,$possibility[0][$shortest],$possibility[0][$average];
			print OUT "$possibility[0][$protonA_dim] --- $possibility[0][$protonB_dim]\n";
			last CHECK;
		}
	}
} else {
	print OUT "$message Not writing restraint.\n";
	$unassigned_xpks .=  "#$xpkno ";
}
}

sub UPDATE_PEAK_LIST
{
# if there is at least one possibility assume first possibility is the assignment
# then check all the other possibilities to make sure the assignment for
# that dimension is the same.  If not the same then label peak "null"
if ($number_possibilities > 0) {
	foreach $dimension (1..$number_dimensions) {
		$output_ass[$dimension] = $possibility[0][$dimension];
		foreach $possibility_number (1..$#possibility) {
			if ( $possibility[$possibility_number][$dimension] ne $output_ass[$dimension] ) {
				$output_ass[$dimension] = "null";
				last;
			}
		}
	}
} else {
	foreach $dimension (1..$number_dimensions) {
		$output_ass[$dimension] = "null";
	}
}
# if dimension has only one assignment then update assignment in hash xpk_data
foreach $dimension (1..$number_dimensions) {
	if ( $output_ass[$dimension] ne "null" ) {
		if ( $data_type eq "NMRView" ) {
			($residue, $atom) = split(':',lc($output_ass[$dimension]));
			($residue_name, $residue_number) = split('_',$residue);
			$xpk_data{$xpkno}[$dimension*7 - 6] = "$residue_number.$atom";
		} elsif ( $data_type eq "tsriNMRView" ) {
			($residue, $atom) = split(':',lc($output_ass[$dimension]));
			($residue_name, $residue_number) = split('_',$residue);
			$xpk_data{$xpkno}[$dimension*6 - 5] = "$residue_number.$atom";
		} elsif ( $data_type eq "Felix" ) {
			$xpk_data{$xpkno}[$dimension*4] = $output_ass[$dimension];
		}
	}
}
}

sub WRITE_NEW_XPK
{
open (NEWXPK, "> $new_xpk");
print NEWXPK "$new_xpk_header";
foreach $xpkno ( sort numerically keys %xpk_data ) {
    if ( $data_type eq "NMRView" ) {
        print NEWXPK "$xpk_data{$xpkno}[0]";
        foreach $dimension (1..$number_dimensions) {
            print NEWXPK "  $xpk_data{$xpkno}[$dimension*7 - 6] ";
            print NEWXPK "  $xpk_data{$xpkno}[$dimension*7 - 5] ";
            print NEWXPK "  $xpk_data{$xpkno}[$dimension*7 - 4] ";
            print NEWXPK "  $xpk_data{$xpkno}[$dimension*7 - 3] ";
            print NEWXPK "  $xpk_data{$xpkno}[$dimension*7 - 2] ";
            print NEWXPK "  $xpk_data{$xpkno}[$dimension*7 - 1] ";
            print NEWXPK "  $xpk_data{$xpkno}[$dimension*7] ";
        }
        print NEWXPK " $xpk_data{$xpkno}[$number_dimensions*7 + 1]";
        print NEWXPK " $xpk_data{$xpkno}[$number_dimensions*7 + 2]";
        print NEWXPK " $xpk_data{$xpkno}[$number_dimensions*7 + 3]";
        print NEWXPK " $xpk_data{$xpkno}[$number_dimensions*7 + 4]";
        print NEWXPK " $xpk_data{$xpkno}[$number_dimensions*7 + 5]\n";
	} elsif ( $data_type eq "tsriNMRView" ) {
		print NEWXPK "$xpk_data{$xpkno}[0]";
		foreach $dimension (1..$number_dimensions) {
			print NEWXPK "  $xpk_data{$xpkno}[$dimension*6 - 5] ";
			print NEWXPK "  $xpk_data{$xpkno}[$dimension*6 - 4] ";
			print NEWXPK "  $xpk_data{$xpkno}[$dimension*6 - 3] ";
			print NEWXPK "  $xpk_data{$xpkno}[$dimension*6 - 2] ";
			print NEWXPK "  $xpk_data{$xpkno}[$dimension*6 - 1] ";
			print NEWXPK "  $xpk_data{$xpkno}[$dimension*6] ";
		}
		print NEWXPK " $xpk_data{$xpkno}[$number_dimensions*6 + 1]";
		print NEWXPK " $xpk_data{$xpkno}[$number_dimensions*6 + 2]";
		print NEWXPK " $xpk_data{$xpkno}[$number_dimensions*6 + 3]\n";
	} elsif ( $data_type eq "Felix" ) {
		printf NEWXPK "%6d", $xpk_data{$xpkno}[0];
		foreach $dimension (1..$number_dimensions) {
			printf NEWXPK " %9.3f", $xpk_data{$xpkno}[$dimension*4 - 3];
			printf NEWXPK " %9.3f", $xpk_data{$xpkno}[$dimension*4 - 2];
			printf NEWXPK " %5d ", $xpk_data{$xpkno}[$dimension*4 - 1];
			printf NEWXPK " %-15s ", $xpk_data{$xpkno}[$dimension*4];
		}
		printf NEWXPK "%10.2f\n", $xpk_data{$xpkno}[$number_dimensions*4 + 1];
	}
}
close (NEWXPK);
}

sub WRITE_UNASS_XPK
{
open (UNASS, "> $unass_xpk");
print UNASS "$new_xpk_header";
foreach $xpkno ( sort numerically keys %xpk_data ) {
	if ($unassigned_xpks =~ /\#$xpkno\s/) {
        if ( $data_type eq "NMRView" ) {
            print UNASS "$xpk_data{$xpkno}[0]";
            foreach $dimension (1..$number_dimensions) {
                print UNASS "  $xpk_data{$xpkno}[$dimension*7 - 6] ";
                print UNASS "  $xpk_data{$xpkno}[$dimension*7 - 5] ";
                print UNASS "  $xpk_data{$xpkno}[$dimension*7 - 4] ";
                print UNASS "  $xpk_data{$xpkno}[$dimension*7 - 3] ";
                print UNASS "  $xpk_data{$xpkno}[$dimension*7 - 2] ";
                print UNASS "  $xpk_data{$xpkno}[$dimension*7 - 1] ";
                print UNASS "  $xpk_data{$xpkno}[$dimension*7] ";
            }
            print UNASS " $xpk_data{$xpkno}[$number_dimensions*7 + 1]";
            print UNASS " $xpk_data{$xpkno}[$number_dimensions*7 + 2]";
            print UNASS " $xpk_data{$xpkno}[$number_dimensions*7 + 3]";
            print UNASS " $xpk_data{$xpkno}[$number_dimensions*7 + 4]";
            print UNASS " $xpk_data{$xpkno}[$number_dimensions*7 + 5]\n";
		} elsif ( $data_type eq "tsriNMRView" ) {
			print UNASS "$xpk_data{$xpkno}[0]";
			foreach $dimension (1..$number_dimensions) {
				print UNASS "  $xpk_data{$xpkno}[$dimension*6 - 5] ";
				print UNASS "  $xpk_data{$xpkno}[$dimension*6 - 4] ";
				print UNASS "  $xpk_data{$xpkno}[$dimension*6 - 3] ";
				print UNASS "  $xpk_data{$xpkno}[$dimension*6 - 2] ";
				print UNASS "  $xpk_data{$xpkno}[$dimension*6 - 1] ";
				print UNASS "  $xpk_data{$xpkno}[$dimension*6] ";
			}
			print UNASS " $xpk_data{$xpkno}[$number_dimensions*6 + 1]";
			print UNASS " $xpk_data{$xpkno}[$number_dimensions*6 + 2]";
			print UNASS " $xpk_data{$xpkno}[$number_dimensions*6 + 3]\n";
		} elsif ( $data_type eq "Felix" ) {
			printf UNASS "%6d", $xpk_data{$xpkno}[0];
			foreach $dimension (1..$number_dimensions) {
				printf UNASS " %9.3f", $xpk_data{$xpkno}[$dimension*4 - 3];
				printf UNASS " %9.3f", $xpk_data{$xpkno}[$dimension*4 - 2];
				printf UNASS " %5d ", $xpk_data{$xpkno}[$dimension*4 - 1];
				printf UNASS " %-15s ", $xpk_data{$xpkno}[$dimension*4];
			}
			printf UNASS "%10.2f\n", $xpk_data{$xpkno}[$number_dimensions*4 + 1];
		}
	}
}
close(UNASS);
}
